
//*******************************************************************
//*																	*
//* Virtual Sample Demo Synthesizer PlugIn							*
//*																	*
//*																	*	
//*																	*	
//*	This sample code emulates a very simple synthesizer with 2   	*	
//*	oscilators and 2 waveforms for each oscillator.  				*	
//* It should give you an idea how to write a Virtual Sampler       *
//* Synthesizer plugin.  											*	
//*																	*	
//*******************************************************************

#include "synthdemo.h"


//*******************************************************************
//*																	*
//* Member Functions												*
//*																	*
//*																	*	
//*******************************************************************

//Frequency table in Hz for 0..128 MIDI notes
static bool FreqTableInitialized=false;
#define FreqTableSize 128 //128 MIDI notes
static float FreqTable[FreqTableSize];


//WaveForms tables
static bool WaveFormsInitialized=false;
#define WaveTableSize 2048
static float PulseTable[WaveTableSize];
static float SawToothTable[WaveTableSize];

//Constructor - setup
VSDemoSynth::VSDemoSynth(int ASampleRate)
		: VSSynthBase(ASampleRate) 			  
{
 //Initialize global tables
 if (!FreqTableInitialized)
 {
  // build frequency table
  double k=1.059463094359; // 12th root of 2
  double a=6.875;
  a*=k;	
  a*=k;	
  a*=k;	//Start with Note 0 (C)
  for (int t=0;t<FreqTableSize;t++)	
  {
   FreqTable[t]=(float)a;
   a*=k; 	
  }
  FreqTableInitialized=true;
 }

 if (!WaveFormsInitialized)
 {
  // build waveforms tables
  for (int t=0;t<WaveTableSize;t++)
  {
   SawToothTable[t]=(float)(-1.0+(2.0*((double)t/(double)WaveTableSize)));
   PulseTable[t]=(t<(int)(WaveTableSize/4))?-1.0f:1.0f; //1:3 pulse
  }
  WaveFormsInitialized=true;
 }

 InitializeProgram();
}

//Destructor - cleanup
VSDemoSynth::~VSDemoSynth()
{
}


//returns count of programs
int VSDemoSynth::GetProgramCount()
{
 return 0; //we don't support programs, otherwise return the number of programs
}

//returns program name
void VSDemoSynth::GetProgramName(int Index,char *Result)
{
 //returns a particular program name, Index ranges from 0..ProgramCount-1
 //since we don't support programs, we leave this function empty
}

//sets program
void VSDemoSynth::SetProgram(int NewValue)
{
 //sets a particular program, Index ranges from 0..ProgramCount-1
 //since we don't support programs, we leave this function empty
}

//returns group count
int VSDemoSynth::GetGroupCount()
{
 //Synthesizer parameters are divided into groups.
 //Each group can have up to 16 parameters
 return 3; //we have 3 groups (General,Oscillator1 and Oscillator2)
}

//returns group name
void VSDemoSynth::GetGroupName(int Index,char *Result)
{
 //returns a particular group name, Index ranges from 0..GroupCount-1
 switch(Index)
 {
  case 0:strcpy(Result,"General");break;
  case 1:strcpy(Result,"Oscillator 1");break;
  case 2:strcpy(Result,"Oscillator 2");break;
  default:Result[0]=0;
 }
}

//returns parameter count
int VSDemoSynth::GetParamCount(int GroupIndex)
{
 switch(GroupIndex)
 {
  case 0:return 1;break; 
  case 1: case 2:return 3;break;
  default:return 0;
 }	
}

//returns label for a parameter
void VSDemoSynth::GetParamLabel(int GroupIndex,int Index,char *Result)
{
 //returns a label string that is displayed for a particular parameter
 //GroupIndex ranges from 0..GroupCount-1.
 //Index ranges from 0..ParamCount-1
 switch(GroupIndex)
 {
  case 0: //General
  {
   switch(Index)
   {
    case 0:strcpy(Result,"Gain");break;
   }
   break;
  } 
  case 1: //Oscillator1
  {	
   switch(Index)
   {
	case 0:strcpy(Result,"Osc1 Wave");break;
	case 1:strcpy(Result,"Osc1 Det");break;
	case 2:strcpy(Result,"Osc1 Vol");break;
   }
   break;
  }
  case 2: //Oscillator 2
  {
   switch(Index)
   {
    case 0:strcpy(Result,"Osc2 Wave");break;
	case 1:strcpy(Result,"Osc2 Det");break;	
    case 2:strcpy(Result,"Osc2 Vol");break;
   }	   
   break;
  }
 }
}

//returns display for a parameter
void VSDemoSynth::GetParamDisplay(int GroupIndex,int Index,char *Result)
{
 //returns a display string that is displayed for the current value of a
 //particular parameter.
 //GroupIndex ranges from 0..GroupCount-1.
 //Index ranges from 0..ParamCount-1
 switch(GroupIndex)
 {
  case 0: //General
  {
   switch(Index)
   {
    case 0:strcpy(Result,Float2String(MasterGain*2));break;
   }
   break;
  } 
  case 1: //Oscillator1
  {
   switch(Index)
   {
    case 0:
	{
	 if (Osc1WaveForm<0.5f) strcpy(Result,"Saw");
	 else strcpy(Result,"Pulse");
     break;
    }
    case 1:
	{
	 strcpy(Result,Int2String((int)(Osc1Detune*2400-1200))); //from -1200..1200 cents
	 strcat(Result," cents");
	 break;
    }
	case 2:
	{
	 strcpy(Result,Int2String((int)(Osc1Volume*100)));
	 strcat(Result," %");
	 break;
    }
   }
   break;
  }
  case 2: //Oscillator2
  {
   switch(Index)
   {
    case 0:
	{
     if (Osc2WaveForm<0.5f) strcpy(Result,"Saw");
	 else strcpy(Result,"Pulse");
     break;
    } 
    case 1:
	{		
	 strcpy(Result,Int2String((int)(Osc2Detune*2400-1200))); //from -1200..1200 cents
	 strcat(Result," cents");
	 break;
	}
	case 2:
	{
	 strcpy(Result,Int2String((int)(Osc2Volume*100)));
	 strcat(Result," %");
	 break;
    }
   } 
   break;
  }
 }
}

//gets value of a parameter
void VSDemoSynth::GetParamValue(int GroupIndex,int Index,float *Result)
{
 //Parameter values range from -1..1
 //GroupIndex ranges from 0..GroupCount-1.
 //Index ranges from 0..ParamCount-1
 switch(GroupIndex)
 {
  case 0: //General
  {
   switch(Index)
   {
    case 0:*Result=MasterGain;break;
   }
   break;
  }
  case 1: //Oscillator1
  {
   switch(Index)
   {
    case 0:*Result=Osc1WaveForm;break;
    case 1:*Result=Osc1Detune;break;
	case 2:*Result=Osc1Volume;break;
   }
   break;
  }
  case 2: //Oscillator 2
  {
   switch(Index)
   {
	case 0:*Result=Osc2WaveForm;break;
    case 1:*Result=Osc2Detune;break;
	case 2:*Result=Osc2Volume;break;	
   }	   
   break;
  }
 }
}

//sets value of a parameter
void VSDemoSynth::SetParamValue(int GroupIndex,int Index,float NewValue)
{
 //Parameter values range from -1..1
 //GroupIndex ranges from 0..GroupCount-1.
 //Index ranges from 0..ParamCount-1
 switch(GroupIndex)
 {
  case 0: //General
  {
   switch(Index)
   {
    case 0:MasterGain=NewValue;break; 
   }
   break;
  }
  case 1: //Oscillator1
  {
   switch(Index)
   { 
    case 0:Osc1WaveForm=NewValue;break;
    case 1:Osc1Detune=NewValue;break;
	case 2:Osc1Volume=NewValue;break;
   }
   break;
  }
  case 2: //Oscillator2
  {
   switch(Index)
   {
    case 0:Osc2WaveForm=NewValue;break;
    case 1:Osc2Detune=NewValue;break;
	case 2:Osc2Volume=NewValue;break;	
   }
   break;
  } 
 }
}


//returns size of Parameter record
void VSDemoSynth::GetParamsInfo(int *ParamsLen)
{
 //returns the size of the buffer that is needed to store the
 //parameters of the synth. This value is used to calculate the
 //buffer size when calling "StoreParams"
 *ParamsLen=7*sizeof(float); //we have 7 parameters
}


//stores parameters into buffer
void VSDemoSynth::StoreParams(void *Params,int ParamsLen)
{
 //The BufferSize passed matches the return value from "GetParamsInfo"
 if (ParamsLen!=7*sizeof(float)) return; //Parameter size not accepted
 ((float *)Params)[0]=Osc1WaveForm;
 ((float *)Params)[1]=Osc1Detune;
 ((float *)Params)[2]=Osc1Volume;
 ((float *)Params)[3]=Osc2WaveForm;
 ((float *)Params)[4]=Osc2Detune;
 ((float *)Params)[5]=Osc2Volume;
 ((float *)Params)[6]=MasterGain;
}


//loads parameters from buffer
void VSDemoSynth::LoadParams(void *Params,int ParamsLen)
{
 //The BufferSize passed matches the return value from "GetParamsInfo"
 if (ParamsLen!=7*sizeof(float)) return; //Parameter size not accepted
 Osc1WaveForm=((float *)Params)[0];
 Osc1Detune=((float *)Params)[1];
 Osc1Volume=((float *)Params)[2];
 Osc2WaveForm=((float *)Params)[3];
 Osc2Detune=((float *)Params)[4];
 Osc2Volume=((float *)Params)[5];
 MasterGain=((float *)Params)[6];
}

//Initializes current program with default values
void VSDemoSynth::InitializeProgram()
{
 //we don't support any programs, reset global parameters
 Osc1WaveForm=0.0f; //saw
 Osc1Detune=0.5f;
 Osc1Volume=0.33f;

 Osc2WaveForm=1.0f; //pulse
 Osc2Detune=0.5f;
 Osc2Volume=0.33f;
 
 MasterGain=0.4f;
}

//returns the file extension used to load and store files for this synth.
//The string returned contains a three letter file extension, used to store
//and load parameter files (eg. "syn"). If the synth supports more than
//one file format, the extensions are separated by "|", e.g. "syn|dx7"
void VSDemoSynth::GetFileExt(char *Result)
{
 //we don't support load/save
 Result[0]=0;
}

//Save the synth parameters to the filename passed. If the function
//returns an error (-1), the error text can be copied to "FileName"
int VSDemoSynth::SaveProgram(char *FileName)
{
 strcpy(FileName,"This synth does not support loading/saving");
 return -1; //we don't support load/save
}


//Gets the Programs and Parameter lens for the given filename.
//Normally you can set "ProgramCount" to 1. If a given file contains more
//than one synthesizer program (as for the DX7 emulator of VSampler), you
//can assign a value>1 here.
//The "ParamsLen" argument must be altered with the parameter buffer len
//for each program. This information is used to allocate the buffers when
//calling "LoadProgram" (see below)
void VSDemoSynth::GetProgramInfo(char *FileName,int *ProgramCount,int *ParamsLen)
{
 //we don't support loading/saving
 *ProgramCount=0;
 *ParamsLen=0;
}

//Loads the synth parameters to the filename passed. If the function
//returns an error (-1), the error text can be copied to "FileName".
//Otherwise the "Result" parameter is an array of TSynthProgramInfo
//records. This array ranges from 0..ProgramCount (returned from "GetProgramInfo").
//Each record has a parameter pointer that was allocated by the caller with
//the "ParamsLen" return value from "GetProgramInfo". You should not reassign
//or free this pointer !
int VSDemoSynth::LoadProgram(char *FileName,PSynthProgramInfo Result)
{
 strcpy(FileName,"This synth does not support loading/saving");
 return -1; //we don't support load/save
}


//This function is called each time a MIDI Note-On event arrives
//"Velocity" is the velocity of the note played and can be used for realtime
//control of synth parameters
//"ScaleCents" is the scaling of the note in cents that should be added
//to the note frequency (100 cents=1 semitone, 1200 cents=1 octave)
//The values passed must be stored for later calls of the "GetBuffer"
//method when Virtual Sampler actually passes data to the soundcard
void VSDemoSynth::NoteOn(int Note,int Velocity,int ScaleCents)
{
 MidiNote=Note;
 
 MidiVelocity=Velocity;
 MidiScaleCents=ScaleCents;
 
 //calculate delta cents
 int CentsOsc1=ScaleCents+(int)(Osc1Detune*2400-1200);
 int CentsOsc2=ScaleCents+(int)(Osc2Detune*2400-1200);

 int NoteOsc1=Note+CentsOsc1/100;
 int NoteOsc2=Note+CentsOsc2/100;
 
 CentsOsc1%=100;
 CentsOsc2%=100;

 //Calculate base frequencies for each oscillator
 FreqOsc1=FreqTable[NoteOsc1&127];
 double Freq1Osc1=(CentsOsc1<0)?FreqTable[(NoteOsc1-1)&127]:FreqTable[(NoteOsc1+1)&127];
 FreqOsc1+=((Freq1Osc1-FreqOsc1)*CentsOsc1)/100;
 
 FreqOsc2=FreqTable[NoteOsc2&127];
 double Freq1Osc2=(CentsOsc2<0)?FreqTable[(NoteOsc2-1)&127]:FreqTable[(NoteOsc2+1)&127];
 FreqOsc2+=((Freq1Osc2-FreqOsc2)*CentsOsc2)/100;

 //scale the frequencies according to the current SampleRate
 double Scale=(float)((double)WaveTableSize/SampleRate);
 FreqOsc1*=Scale;
 FreqOsc2*=Scale;

 //calculate volume depending on current Velocity
 Volume = (double)(MasterGain * (double)Velocity*(1.0f/127.0f));

 //setup processing
 PhaseOsc1=0.0;
 PhaseOsc2=0.0;
 MidiNoteIsOn=true;
}

//This function is called each time a MIDI Note-Off event arrives
void VSDemoSynth::NoteOff()
{
 MidiNoteIsOn=false;
}

//This is probably the most important function since it does the actual
//sample processing work. The buffer is an array of float values 0..SampleCount-1.
//The sample range is -1..1. This function returns the sample amount copied
//to the buffer. This function replaces the samples in the buffer passed by the caller

//Note: This function is called each time Virtual Sampler sends a buffer to
//      the soundcard and therefore should be optimized for speed
#pragma optimize("agtw",on) //optimize agressively for time
int VSDemoSynth::GetBuffer(float *Buffer,int SampleCount)
{
 if (!MidiNoteIsOn)
 {
  memset(Buffer,0,SampleCount*sizeof(float));
  return 0; //we have done
 }

 int Frames=SampleCount;
 float* wave1=(Osc1WaveForm<0.5)?SawToothTable:PulseTable;
 float* wave2=(Osc2WaveForm<0.5)?SawToothTable:PulseTable;
 while (--Frames>=0)
 {
  (*Buffer)=(float)(wave1[(long)PhaseOsc1&(WaveTableSize-1)]*Osc1Volume*Volume);
  (*Buffer++)+=(float)(wave2[(long)PhaseOsc2&(WaveTableSize-1)]*Osc2Volume*Volume);
  PhaseOsc1+=FreqOsc1;
  PhaseOsc2+=FreqOsc2;
 }

 return SampleCount; //we have processed all samples
}
#pragma optimize("",on) //revert to default optimization options

//*******************************************************************
//*																	*
//* Exported Functions												*
//*																	*
//*																	*	
//*******************************************************************


//Note: The following three functions must always be exported by
//      a Virtual Sampler Synth PlugIn. Otherwise the Synth will not
//      be recognized by Virtual Sampler, These three functions must
//      also be part of your .def file !

void EXPENTRY VSSynthGetInfo(TPlugInSynth *Params)
{
 //you always must copy the default parameters (defined in 
 //vssynthbase.h to the record passes by the caller
 //You only should modify the name of the default record and
 //create a class derived from VSSynthBase for your Synth.
 memcpy(Params,&SynthParams,sizeof(TPlugInSynth));

 //modify the name of the Synth
 strcpy(Params->Name,"VS PlugIn Demo Synth");
}

//create a synthesizer Instance 
VSSynthBase * EXPENTRY VSSynthCreateInstance(int SampleRate)
{
 return new VSDemoSynth(SampleRate);
}

//destroys an instance
void EXPENTRY VSSynthDestroyInstance(VSSynthBase *Instance)
{
 delete Instance;	 
}
